# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/99_Image_resize.ipynb (unless otherwise specified).

__all__ = ['display_image', 'glob_suffixes', 'resize_dataset']

# Cell

from pathlib import Path
from PIL import Image
from typing import Tuple, Optional, Dict, List, Union, Callable, Iterable

# Internal Cell

from fastcore.script import *

#import dolphins

import numpy as np
from shutil import rmtree

from .helpers import *
from tempfile import TemporaryDirectory

# Internal Cell


def _get_pallete(img_P: Image.Image) -> List[List[int]]:
    """Returns a list of colors from the input image"""
    palette = np.array(img_P.getpalette()).reshape((-1, 3)).tolist()
    palette = palette[: len(np.unique(palette, axis=0))]

    return palette

# Cell


def display_image(img: Union[Image.Image, Iterable[Image.Image]], x_size=240):
    """Displays an image or images with smaller dimension equal to x_size,
    and the other dimension is calculated"""

    if isinstance(img, Image.Image):
        y_size = int(img.size[1] / img.size[0] * x_size)
        display(img.resize((x_size, y_size)))
    else:
        for i in img:
            display_image(i, x_size=x_size)

# Internal Cell


def _separate_img(
    img: Image.Image,  # png image
    palette: List[List[int]],  # list of RGB colors
) -> List[Image.Image]:
    """Creates a list of images from the input image and color palette.
    Each image has one class and background"""

    img_array = np.array(img)
    img_separated = [
        Image.fromarray(img_array == i, mode="P") for i, color in enumerate(palette)
    ]

    black = [0, 0, 0]
    for color, im in zip(palette, img_separated):
        im.putpalette(black + color)

    return img_separated

# Internal Cell


def separate_masks(img: Image.Image) -> List[Image.Image]:
    """Converts RGB image to P. Creates a list of images from the input image.
    Each image has one class and background"""

    img_P = img.convert("P", palette=Image.ADAPTIVE, colors=len(img.getcolors()))

    palette = _get_pallete(img_P)

    return _separate_img(img_P, palette)

# Internal Cell


def resize_composite_multi(img: Image.Image, image_size: Tuple) -> Image.Image:
    """Resizes images and merges images into one"""
    separated_masks = separate_masks(img)
    separated_masks_new = []
    for im in separated_masks:
        separated_masks_new.append(im.resize(image_size))

    temp_img = Image.composite(
        separated_masks_new[1].convert("RGB"),
        separated_masks_new[0].convert("RGB"),
        separated_masks_new[1].convert("L"),
    )
    if len(separated_masks_new) <= 3:
        return temp_img
    else:
        for k in range(2, len(separated_masks_new) - 1):
            temp_img = Image.composite(
                separated_masks_new[k].convert("RGB"),
                temp_img.convert("RGB"),
                separated_masks_new[k]
                .convert("L")
                .point(lambda x: 0 if x == 0 else 255, "1"),
            )
        return temp_img

# Internal Cell


def calculate_image_size(img: Image.Image, smaller_size=1000) -> Tuple[int]:
    """Sets the smaller dimension of the image to the value of the variable smaller_size and reduces the larger dimension by the same amount"""
    x, y = img.size

    if min(x, y) == y:

        x = int(x / (y / smaller_size))
        y = smaller_size
    else:
        y = int(y / (x / smaller_size))
        x = smaller_size

    return (x, y)

# Internal Cell


def file_suffix_in(f: Path, suffixes: List[str]) -> bool:
    """Checks if the path contains a suffix"""
    return f.suffix[1:].lower() in [suffix.lower() for suffix in suffixes]

# Cell


def glob_suffixes(root_path: Path, suffixes: Union[List[str], str]) -> List[Path]:
    """Returns all suffixes located in the path and in the input variable suffixes"""
    if isinstance(suffixes, str):
        suffixes = [suffixes]
    return sorted([f for f in root_path.glob("**/*") if file_suffix_in(f, suffixes)])

# Internal Cell


def get_files_from_path(src_path: Path) -> (Path, Path, Path):
    """Returns three lists of files from the input directory"""
    src_images_path = Path(src_path / "JPEGImages/")
    src_class_path = Path(src_path / "SegmentationClass/")
    src_instance_path = Path(src_path / "SegmentationObject/")

    ensure_exists(src_images_path)
    ensure_exists(src_class_path)
    ensure_exists(src_instance_path)

    src_images = glob_suffixes(src_images_path, ("jpg", "jpeg"))
    src_lbl_classes = glob_suffixes(src_class_path, "png")
    src_lbl_instances = glob_suffixes(src_instance_path, "png")

    return src_images, src_lbl_classes, src_lbl_instances

# Internal Cell


def recreate_dst_dir(dst_path: Path) -> (Path, Path, Path):
    """Deletes and recreates the directory and returns 3 dst paths"""
    dst_images_path = Path(dst_path / "JPEGImages")
    dst_class_path = Path(dst_path / "SegmentationClass")
    dst_instance_path = Path(dst_path / "SegmentationObject")

    rmtree(dst_images_path, ignore_errors=True)
    rmtree(dst_class_path, ignore_errors=True)
    rmtree(dst_instance_path, ignore_errors=True)

    dst_images_path.mkdir(parents=True, exist_ok=True)
    dst_class_path.mkdir(parents=True, exist_ok=True)
    dst_instance_path.mkdir(parents=True, exist_ok=True)

    return dst_images_path, dst_class_path, dst_instance_path

# Internal Cell


def save_img_lbl_inst(
    src_img_file,
    src_class_file,
    src_instance_file,
    dst_images_path,
    dst_class_path,
    dst_instance_path,
    smaller_size=1000,
):
    """Resizes images from the src directories and saves them to the dst directories"""

    lbl = Image.open(src_class_file)
    image_size = calculate_image_size(lbl, smaller_size)
    lbl = resize_composite_multi(lbl, image_size)

    inst = Image.open(src_instance_file)
    inst = resize_composite_multi(inst, image_size)

    img = Image.open(src_img_file)
    img = img.resize(image_size)

    dst_img_file = dst_images_path / src_img_file.name
    dst_class_file = dst_class_path / src_class_file.name
    dst_instance_file = dst_instance_path / src_instance_file.name

    img.save(dst_img_file)
    lbl.save(dst_class_file)
    inst.save(dst_instance_file)

# Cell


@call_parse
def resize_dataset(
    src_path: Param("input directory containing images and annotations", Path),
    dst_path: Param("output directory, will be deleted if not empy", Path),
    dst_image_size: Param("min height or width", int) = 1000,
    verbose: Param("", bool) = True,
):

    """Resizes the entire dataset (images and labels) and saves it"""

    src_images, src_lbl_classes, src_lbl_instances = get_files_from_path(src_path)

    for src_img_path, src_class_path, src_instance_path in zip(
        src_images, src_lbl_classes, src_lbl_instances
    ):
        ensure_equal(src_img_path.stem, src_class_path.stem)
        ensure_equal(src_img_path.stem, src_instance_path.stem)

    dst_images_path, dst_class_path, dst_instance_path = recreate_dst_dir(dst_path)

    n = len(src_images)
    print(f"Resizing images and annotation masks in: {src_path.resolve()}")
    for i, (src_img_file, src_class_file, src_instance_file) in enumerate(
        zip(src_images, src_lbl_classes, src_lbl_instances)
    ):
        if verbose:
            print(f" - [{i+1}/{n}] {src_img_file.name}...")

        save_img_lbl_inst(
            src_img_file,
            src_class_file,
            src_instance_file,
            dst_images_path,
            dst_class_path,
            dst_instance_path,
        )


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/04_Leaderboard.ipynb (unless otherwise specified).

__all__ = ['get_leaderboard']

# Internal Cell

import pandas as pd
import numpy as np
import boto3
from pathlib import Path

# Internal Cell

import zipfile
import shutil
import torch
import tempfile

from .datasets import get_dataset
from .instance_segmentation.model import *

# Internal Cell


def parse_filename(fname):
    tmp = fname.split("-")
    date = pd.to_datetime(tmp[1] + tmp[2] + tmp[3])
    alias = tmp[6]
    email = tmp[7]
    submitted_iou = tmp[5].split("=")[1]

    return {
        "file_name": fname,
        "date": date,
        "alias": alias,
        "email": email,
        "submitted_iou": submitted_iou,
        "calculated_iou": np.nan,
    }

# Internal Cell

s3 = boto3.resource("s3")
my_bucket = s3.Bucket("ai-league.cisex.org")
private_leaderboard_path = Path("private_leaderboard.csv")
public_leaderboard_path = Path("leaderboard.csv")


def get_submissions_from_s3(private_leaderboard_path=private_leaderboard_path):
    """Downloads the zip file from s3 if there is no record of it in the csv file"""
    if private_leaderboard_path.exists():
        private_leaderboard = pd.read_csv(private_leaderboard_path)
    else:
        private_leaderboard = dict(file_name=[])

    # download file into models_for_evaluation directory
    s3_objects = [
        s3_object
        for s3_object in my_bucket.objects.all()
        if Path(s3_object.key).match("*submission*.zip")
        and Path(s3_object.key).name not in list(private_leaderboard["file_name"])
    ]
    if len(s3_objects) > 0:
        for i, s3_object in enumerate(s3_objects):
            print(f"Downloading {i+1}/{len(s3_objects)} from S3...")
            my_bucket.download_file(s3_object.key, f"models_for_evaluation/{Path(s3_object.key).name}")

        # return new entries
        new_entries = pd.Series([Path(s3_object.key).name for s3_object in s3_objects]).apply(parse_filename).apply(pd.Series)
    else:
        x = "uploaded-2020-12-22T15:35:15.513570-submission-iou=0.46613-dolphin123-name.surname@gmail.com-2020-12-22T15:35:04.875962.zip"
        new_entries = pd.Series([x]).apply(parse_filename).apply(pd.Series).iloc[:0, :]

    return new_entries


# Internal Cell

def public(private_leaderboard):
    return private_leaderboard[["alias", "date", "submitted_iou", "calculated_iou"]]

# Internal Cell

def merge_with_private_leaderboard(
    new_entries, private_leaderboard_path=private_leaderboard_path
):
    # merge private leaderboard and new_entries if needed
    new_entries["calculated_iou"] = np.nan
    if private_leaderboard_path.exists():
        private_leaderboard = pd.read_csv(private_leaderboard_path)
        private_leaderboard = pd.concat([private_leaderboard, new_entries], axis=0)
        private_leaderboard = private_leaderboard.drop_duplicates(subset="file_name")
    else:
        private_leaderboard = new_entries

    private_leaderboard.to_csv(private_leaderboard_path, index=False)

    return private_leaderboard

# Internal Cell

def evaluate_model(model_path) -> float:
    # do it
    with tempfile.TemporaryDirectory() as d:
        with zipfile.ZipFile(model_path, "r") as zip_ref:
            zip_ref.extractall(path=d)
            unzipped_path = [x for x in Path(d).glob("submiss*")][0]

        model = torch.load(unzipped_path / "model.pt")
        data_loader, data_loader_test = get_dataset("segmentation", batch_size=4)
        iou, iou_df = iou_metric(model, data_loader_test.dataset)

    return iou

# Internal Cell

def evaluate_private_leaderboard(private_leaderboard_path=private_leaderboard_path):
    private_leaderboard = pd.read_csv(private_leaderboard_path)
    new_entries = private_leaderboard.loc[private_leaderboard["calculated_iou"].isna()]

    n = new_entries.shape[0]
    for i, ix in enumerate(new_entries.index):
        row = new_entries.loc[ix]
        file_name, alias, dt = row["file_name"], row["alias"], row["date"]
        print(f"Evaluating model {i+1}/{n} for {alias} submitted at {dt}...")
        calculated_iou = evaluate_model(f"models_for_evaluation/{file_name}")
        private_leaderboard.loc[ix, "calculated_iou"] = calculated_iou

    private_leaderboard.to_csv(private_leaderboard_path, index=False)
    return private_leaderboard

# Internal Cell

def save_public_leaderboard(private_leaderboard_path=private_leaderboard_path, public_leaderboard_path=public_leaderboard_path):
    private_leaderboard = pd.read_csv(private_leaderboard_path)
    public_leaderboard = public(private_leaderboard)
    public_leaderboard.to_csv(public_leaderboard_path, index=False)

# Cell


def get_leaderboard(public_leaderboard_path=public_leaderboard_path):
    public_leaderboard = pd.read_csv(public_leaderboard_path)
    public_leaderboard = public_leaderboard[
        (public_leaderboard.alias != "dolphin123")
        & (public_leaderboard.alias != "malimedo")
        & (public_leaderboard.alias != "prvi_pokušaj")
    ]
    public_leaderboard = public_leaderboard.sort_values(
        by=["submitted_iou"], ascending=False
    ).reset_index(drop=True)
    public_leaderboard.drop_duplicates(subset="alias", keep="first", inplace=True)

    public_leaderboard = public_leaderboard.sort_values(
        by=["submitted_iou"], ascending=False
    ).reset_index(drop=True)
    public_leaderboard.index = public_leaderboard.index + 1
    return public_leaderboard


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"dataset_url": "01_Datasets.ipynb",
         "dataset_root": "01_Datasets.ipynb",
         "dataset_zip": "01_Datasets.ipynb",
         "DolphinsInstanceSegmentationDataset": "01_Datasets.ipynb",
         "ToTensor": "01_Datasets.ipynb",
         "stack_imgs": "01_Datasets.ipynb",
         "display_batches": "01_Datasets.ipynb",
         "get_image2tensor_transforms": "01_Datasets.ipynb",
         "get_dataset": "01_Datasets.ipynb",
         "Compose": "01_Datasets.ipynb",
         "RandomHorizontalFlip": "01_Datasets.ipynb",
         "device": "02_Model.ipynb",
         "train_one_epoch": "02_Model.ipynb",
         "show_prediction": "02_Model.ipynb",
         "show_predictions": "02_Model.ipynb",
         "get_true_and_predicted_masks": "02_Model.ipynb",
         "iou_metric_mask_pair": "02_Model.ipynb",
         "iou_metric_matrix_of_example": "02_Model.ipynb",
         "largest_values_in_row_colums": "02_Model.ipynb",
         "iou_metric_example": "02_Model.ipynb",
         "iou_metric": "02_Model.ipynb",
         "PermutedDataset": "02_Model.ipynb",
         "show_predictions_sorted_by_iou": "02_Model.ipynb",
         "get_presigned_url_from_aws": "03_Submissions.ipynb",
         "upload_with_presigned_url": "03_Submissions.ipynb",
         "submit_model": "03_Submissions.ipynb",
         "parse_filename": "04_Leaderboard.ipynb",
         "get_submissions_from_s3": "04_Leaderboard.ipynb",
         "s3": "04_Leaderboard.ipynb",
         "my_bucket": "04_Leaderboard.ipynb",
         "private_leaderboard_path": "04_Leaderboard.ipynb",
         "public_leaderboard_path": "04_Leaderboard.ipynb",
         "public": "04_Leaderboard.ipynb",
         "merge_with_private_leaderboard": "04_Leaderboard.ipynb",
         "evaluate_model": "04_Leaderboard.ipynb",
         "evaluate_private_leaderboard": "04_Leaderboard.ipynb",
         "save_public_leaderboard": "04_Leaderboard.ipynb",
         "get_leaderboard": "04_Leaderboard.ipynb",
         "SmoothedValue": "09_Utils.ipynb",
         "all_gather": "09_Utils.ipynb",
         "reduce_dict": "09_Utils.ipynb",
         "MetricLogger": "09_Utils.ipynb",
         "collate_fn": "09_Utils.ipynb",
         "warmup_lr_scheduler": "09_Utils.ipynb",
         "mkdir": "09_Utils.ipynb",
         "setup_for_distributed": "09_Utils.ipynb",
         "is_dist_avail_and_initialized": "09_Utils.ipynb",
         "get_world_size": "09_Utils.ipynb",
         "get_rank": "09_Utils.ipynb",
         "is_main_process": "09_Utils.ipynb",
         "save_on_master": "09_Utils.ipynb",
         "init_distributed_mode": "09_Utils.ipynb",
         "ensure": "97_Helpers.ipynb",
         "ensure_equal": "97_Helpers.ipynb",
         "ensure_exists": "97_Helpers.ipynb",
         "fix_readme": "97_Helpers.ipynb",
         "glob_suffixes": "99_Image_resize.ipynb",
         "get_suffixes": "98_Convert_raw_image_to_jpg.ipynb",
         "get_rand_file_with_suff": "98_Convert_raw_image_to_jpg.ipynb",
         "get_one_image_for_each_suffix": "98_Convert_raw_image_to_jpg.ipynb",
         "create_sample_test_dir_if_needed": "98_Convert_raw_image_to_jpg.ipynb",
         "convert_image_to_jpg_darktable": "98_Convert_raw_image_to_jpg.ipynb",
         "convert_files_with_darktable": "98_Convert_raw_image_to_jpg.ipynb",
         "display_image": "99_Image_resize.ipynb",
         "separate_masks": "99_Image_resize.ipynb",
         "resize_composite_multi": "99_Image_resize.ipynb",
         "calculate_image_size": "99_Image_resize.ipynb",
         "file_suffix_in": "99_Image_resize.ipynb",
         "get_files_from_path": "99_Image_resize.ipynb",
         "recreate_dst_dir": "99_Image_resize.ipynb",
         "save_img_lbl_inst": "99_Image_resize.ipynb",
         "resize_dataset": "99_Image_resize.ipynb"}

modules = ["datasets.py",
           "instance_segmentation/model.py",
           "submissions.py",
           "leaderboard.py",
           "utils.py",
           "helpers.py",
           "convert_raw_jpg.py",
           "image_resize.py"]

doc_url = "https://cro-ai-league.cisex.org/"

git_url = "https://github.com/cro-ai-league/dolphins-recognition-challenge/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/02_Model.ipynb (unless otherwise specified).

__all__ = ['train_one_epoch', 'show_prediction', 'show_predictions', 'iou_metric_mask_pair',
           'iou_metric_matrix_of_example', 'largest_values_in_row_colums', 'iou_metric_example', 'iou_metric',
           'show_predictions_sorted_by_iou']

# Cell

from pathlib import Path
from typing import List, Tuple, Union, Optional, Dict, Set

# Internal Cell

import numpy as np
import pandas as pd
from datetime import datetime
import PIL
from PIL import Image
from zipfile import ZipFile
import random
import math

import torch
import torch.utils.data
from torch.hub import download_url_to_file

import torchvision
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
from torchvision.transforms import ToPILImage

from ..datasets import stack_imgs
from dolphins_recognition_challenge import utils

from ..datasets import get_dataset

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

# Cell

def train_one_epoch(
    model,
    optimizer,
    data_loader,
    device,
    epoch,
    print_freq=10,
):
    """ Trains one epoch of the model. Copied from the reference implementation from https://github.com/pytorch/vision.git.
    """
    model.train()
    metric_logger = utils.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', utils.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)

    lr_scheduler = None
    if epoch == 0:
        warmup_factor = 1. / 1000
        warmup_iters = min(1000, len(data_loader) - 1)

        lr_scheduler = utils.warmup_lr_scheduler(optimizer, warmup_iters, warmup_factor)

    for images, targets in metric_logger.log_every(data_loader, print_freq, header):
        images = list(image.to(device) for image in images)
        targets = [{k: v.to(device) for k, v in t.items()} for t in targets]

        loss_dict = model(images, targets)

        losses = sum(loss for loss in loss_dict.values())

        # reduce losses over all GPUs for logging purposes
        loss_dict_reduced = utils.reduce_dict(loss_dict)
        losses_reduced = sum(loss for loss in loss_dict_reduced.values())

        loss_value = losses_reduced.item()

        if not math.isfinite(loss_value):
            print("Loss is {}, stopping training".format(loss_value))
            print(loss_dict_reduced)
            sys.exit(1)

        optimizer.zero_grad()
        losses.backward()
        optimizer.step()

        if lr_scheduler is not None:
            lr_scheduler.step()

        metric_logger.update(loss=losses_reduced, **loss_dict_reduced)
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])

    return metric_logger


# Cell

def show_prediction(
    model,
    img: torch.Tensor(),
    *,
    score_threshold: float=0.5,
    width: int=820
) -> None:
    """ Show a single prediction by the model
    """
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    # convert Tensor to PIL Image
    img_bg = Image.fromarray(img.mul(255).permute(1, 2, 0).byte().numpy())
    images = [img_bg]

    model.eval()
    with torch.no_grad():
        prediction = model([img.to(device)])
    predicted_masks = prediction[0]["masks"]
    scores = prediction[0]["scores"]

    for i in range(predicted_masks.shape[0]):
        score = scores[i]
        if score >= score_threshold:
            bg = img_bg.copy()
            fg = Image.fromarray(predicted_masks[i, 0].mul(255).byte().cpu().numpy())
            bg.paste(fg.convert("RGB"), (0, 0), fg)
            images.append(bg)

    display(stack_imgs(images, width))


# Cell

def show_predictions(
    model,
    *,
    data_loader=None,
    dataset=None,
    n=None,
    score_threshold=0.5,
    iou_df=None,
    width=820
):
    """ Show at most `n` predictions for examples in a given data loader.
    """
    assert (data_loader is None) ^ (dataset is None), f"only one of dataloader ({dataloader}) and dataset({dataset}) must be defined"

    if data_loader:
        dataset = data_loader.dataset

    if n == None:
        n = len(dataset)
    else:
        n = min(n, len(dataset))

    for i in range(n):
        if iou_df is not None:
            print(f"IOU metric: {iou_df['iou'].iloc[i]}")
        show_prediction(model, img=dataset[i][0], score_threshold=score_threshold, width=width)

# Internal Cell

def get_true_and_predicted_masks(
    model: torchvision.models.detection.mask_rcnn.MaskRCNN,
    example: Tuple[torch.Tensor, Dict[str, torch.Tensor]],
    score_threshold: float = 0.5,
) -> Tuple[PIL.Image.Image, Dict[str, np.array]]:
    """ Returns a PIL image and dictionary containing both true and predicted masks as numpy arrays.
    """

    img = example[0]

    true_masks = (
        example[1]["masks"].mul(255).cpu().numpy().astype(np.int8)
    )

    model.eval()
    with torch.no_grad():
        predictions = model([img.to(device)])

    pred_scores = predictions[0]["scores"].cpu().numpy()

    pred_masks = predictions[0]["masks"].squeeze(1).mul(255).cpu().numpy().astype(np.int8)
    pred_masks = np.squeeze(pred_masks[np.argwhere(pred_scores >= score_threshold), :, :], 1)

    return ToPILImage()(img), {"true": true_masks, "predicted": pred_masks}

# Cell

def iou_metric_mask_pair(
    binary_segmentation: np.array,
    binary_gt_label: np.array,
) -> float:
    """
    Compute the IOU between two binary segmentation (typically one ground truth and a predicted one).
    Input:
        binary_segmentation: binary 2D numpy array representing the region of interest as segmented by the algorithm
        binary_gt_label: binary 2D numpy array representing the region of interest as provided in the database
    Output:
        IOU: IOU between the segmentation and the ground truth
    """

    assert binary_segmentation.dtype in [np.int, np.int8, np.int16, np.int32, np.bool]
    assert binary_gt_label.dtype in [np.int, np.int8, np.int16, np.int32, np.bool]
    assert len(binary_segmentation.shape) == 2
    assert len(binary_gt_label.shape) == 2

    # turn all variables to booleans, just in case
    binary_segmentation = np.asarray(binary_segmentation, dtype=np.bool)
    binary_gt_label = np.asarray(binary_gt_label, dtype=np.bool)

    # compute the intersection
    intersection = np.logical_and(binary_segmentation, binary_gt_label)
    union = np.logical_or(binary_segmentation, binary_gt_label)

    # count the number of True pixels in the binary segmentation
    segmentation_pixels = float(np.sum(binary_segmentation.flatten()))

    # same for the ground truth
    gt_label_pixels = float(np.sum(binary_gt_label.flatten()))

    # same for the intersection and union
    intersection = float(np.sum(intersection.flatten()))
    union = float(np.sum(union.flatten()))

    # compute the Dice coefficient
    smooth = 0.001
    iou = (intersection + smooth) / (union + smooth)

    return iou

# Cell


def iou_metric_matrix_of_example(
    model: torchvision.models.detection.mask_rcnn.MaskRCNN,
    example: Tuple[torch.Tensor, Dict[str, torch.Tensor]],
    score_threshold: float = 0.5
) -> List[List[float]]:
    _, masks = get_true_and_predicted_masks(model, example, score_threshold)

    return np.array(
        [
            [
                iou_metric_mask_pair(
                    binary_segmentation=masks["predicted"][j, :, :],
                    binary_gt_label=masks["true"][i, :, :],
                )
                for i in range(masks["true"].shape[0])
            ]
            for j in range(masks["predicted"].shape[0])
        ]
    )

# Internal Cell

def _argmax2d(xs: np.array) -> Tuple[int, int]:
    assert len(xs.shape) == 2

    n_col = xs.shape[1]
    ij = xs.argmax()
    i = ij // n_col
    j = ij % n_col
    return i, j

def _drop_max_row_and_column(xs: np.array) -> Tuple[float, np.array]:
    i, j = _argmax2d(xs)

    max_value = xs[i, j]

    xs = np.delete(xs, i, 0)
    xs = np.delete(xs, j, 1)

    return max_value, xs

# Internal Cell

def _resize_to_square(xs: np.array) -> np.array:
    new_size = max(xs.shape)
    new_xs = np.zeros((new_size, new_size))
    new_xs[:xs.shape[0], :xs.shape[1]] = xs
    return new_xs

# Cell

def largest_values_in_row_colums(xs: np.array) -> List[float]:
    """ Approximates the largest value in each row/column.
    """
    if xs.shape == (0, ):
        return [0]

    assert len(xs.shape) == 2

    # resize matrix to square dimensions if needed
    if xs.shape[0] != xs.shape[1]:
        xs = _resize_to_square(xs)

    assert xs.shape[0] == xs.shape[1]


    # return the only value if a single value in the matrix
    if xs.shape[0] == 1:
        return [xs[0, 0]]

    # find the largest value in the matirx and recursively find the largest values in the remaining matrix
    max_value, remainder = _drop_max_row_and_column(xs)
    return [max_value] + largest_values_in_row_colums(remainder)

# Cell


def iou_metric_example(
    model: torchvision.models.detection.mask_rcnn.MaskRCNN,
    example: Tuple[torch.Tensor, Dict[str, torch.Tensor]],
    score_threshold: float = 0.5,
) -> float:

    iou_matrix = iou_metric_matrix_of_example(model, example, score_threshold)
    matching_ious = largest_values_in_row_colums(iou_matrix)
    iou = np.mean(matching_ious)

    return iou

# Cell


def iou_metric(
    model: torchvision.models.detection.mask_rcnn.MaskRCNN,
    dataset: torch.utils.data.Dataset,
    score_threshold: float = 0.5,
) -> float:
    """Calculate IOU metric on the whole dataloader"""

    iou = [
        iou_metric_example(model, dataset[i], score_threshold)
        for i in range(len(dataset))
    ]

    img_paths = [f for f in dataset.img_paths]

    iou_df = pd.DataFrame(dict(paths=img_paths, iou=iou)).sort_values(by="iou")

    iou = np.mean(iou)

    return iou, iou_df

# Internal Cell

class PermutedDataset():
    def __init__(self, ds, permutation):
        self.ds = ds
        self.permutation = permutation

    def __getitem__(self, idx):
        return self.ds[self.permutation[idx]]

    def __len__(self):
        return len(self.permutation)

# Cell

def show_predictions_sorted_by_iou(model, dataset):
    iou, iou_df = iou_metric(model, dataset)

    permutation = iou_df.index.to_list()

    sorted_dataset = PermutedDataset(dataset, permutation)

    show_predictions(model, dataset=sorted_dataset, iou_df=iou_df)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/01_Datasets.ipynb (unless otherwise specified).

__all__ = ['ToTensor', 'stack_imgs', 'display_batches', 'get_image2tensor_transforms', 'get_dataset', 'Compose',
           'RandomHorizontalFlip']

# Cell

from pathlib import Path
from typing import *

# Internal Cell


import numpy as np
import shutil
from datetime import datetime
import torch
import torch.utils.data
from torch.hub import download_url_to_file
import torchvision
import PIL
from PIL import Image
from zipfile import ZipFile
import random

from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor

from torchvision.transforms import functional as F
from torchvision.transforms import ToPILImage

from tempfile import TemporaryDirectory

from dolphins_recognition_challenge import utils

import git

# Internal Cell

dataset_url = "https://s3.eu-central-1.amazonaws.com/ai-league.cisex.org/2020-2021/dolphins-instance-segmentation/dolphins_200_train_val.zip"

# Internal Cell

dataset_root = Path("./data/dolphins_200_train_val")
dataset_zip = dataset_root.parent / "dolphins_200_train_val.zip"

def _download_data_if_needed():

    dataset_zip.parent.mkdir(parents=True, exist_ok=True)

    if not dataset_zip.exists():
        torch.hub.download_url_to_file(
            dataset_url,
            dataset_zip,
            hash_prefix=None,
            progress=True,
        )


    with ZipFile(dataset_zip, 'r') as zip_ref:
        zip_ref.extractall(dataset_root)


# Internal Cell


def _enumerate_colors_for_fname(fname: Path) -> Tuple[int, int, int]:
    """Finds all colors in the image"""
    img = Image.open(fname)
    colors = [y for x, y in img.getcolors()]
    return colors

# Internal Cell


def _enumerate_colors_for_fnames(fnames: List[Path]) -> Dict[Tuple[int, int, int], int]:
    """This function is used to pin (0, 0, 0) color to the front of palette"""
    colors = np.array([_enumerate_colors_for_fname(fname) for fname in fnames]).reshape(
        -1, 3
    )
    colors = set([tuple(x) for x in colors.tolist() if tuple(x) != (0, 0, 0)])
    colors = [(0, 0, 0)] + list(colors)
    return {x: i for i, x in enumerate(colors)}

# Internal Cell


def _substitute_values(xs: np.array, x, y):
    """Not sure I understand what this does"""
    ix_x = xs == x
    ix_y = xs == y
    xs[ix_x] = y
    xs[ix_y] = x

# Internal Cell


def _enumerate_image_for_instances(
    im: Image, force_black_to_zero: bool = True, max_colors=16
) -> np.array:
    """convert rgb image mask to enumerated image mask"""
    pallete_mask = im.convert("P", palette=Image.ADAPTIVE, colors=max_colors)

    xs = np.array(pallete_mask)

    if force_black_to_zero:
        _substitute_values(xs, 0, xs.max())

    return xs

# Internal Cell


def _enumerate_image_for_classes(
    im: Image,
    colors: Dict[Tuple[int], int] = None,
) -> np.array:
    """Enumerates classes from the rbg format"""
    xs = np.array(im)
    xs = [
        ((xs == color).all(axis=-1)).astype(int) * code
        for color, code in colors.items()
    ]
    xs_sum = xs[0]
    for i in range(1, len(xs)):
        xs_sum = xs_sum + xs[i]
    return xs_sum.astype("uint8")

# Internal Cell


class DolphinsInstanceSegmentationDataset(torch.utils.data.Dataset):
    """Instance segmentation dataset"""

    def __init__(
        self,
        root: Path,
        tensor_transforms: Optional[Callable[[Image.Image], Any]]=None,
        n_samples: int=-1
    ):
        self.root = root
        self.tensor_transforms = tensor_transforms
        # load all image files, sorting them to
        # ensure that they are aligned
        self.img_paths = sorted((root / "JPEGImages").glob("*.*"))[:n_samples]
        self.label_paths = sorted((root / "SegmentationClass").glob("*.*"))[:n_samples]
        self.mask_paths = sorted((root / "SegmentationObject").glob("*.*"))[:n_samples]

        self.class_colors = _enumerate_colors_for_fnames(self.label_paths)

    def __getitem__(self, idx):

        # load images ad masks
        img_path = self.img_paths[idx]
        label_path = self.label_paths[idx]
        mask_path = self.mask_paths[idx]

        # load and transform images and masks
        img = Image.open(img_path).convert("RGB")
        mask_img = Image.open(mask_path)
        label_img = Image.open(label_path)

        # note that we haven't converted the mask to RGB,
        # because each color corresponds to a different instance
        # with 0 being background
        mask = _enumerate_image_for_instances(mask_img)

        # instances are encoded as different colors
        obj_ids = np.unique(mask)

        # first id is the background, so remove it
        obj_ids = obj_ids[1:]

        # split the color-encoded mask into a set
        # of binary masks
        masks = mask == obj_ids[:, None, None]

        label_array = _enumerate_image_for_classes(label_img, self.class_colors)
        # get bounding box coordinates for each mask
        num_objs = len(obj_ids)
        boxes = []
        labels = []
        for i in range(num_objs):
            pos = np.where(masks[i])
            xmin = np.min(pos[1])
            xmax = np.max(pos[1])
            ymin = np.min(pos[0])
            ymax = np.max(pos[0])
            boxes.append([xmin, ymin, xmax, ymax])

            class_mask = label_array * masks[i]
            label, count = np.unique(class_mask, return_counts=True)
            assert label.shape[0] <= 2
            label = max(label)
            labels.append(label)

        boxes = torch.as_tensor(boxes, dtype=torch.float32)
        # there WAS multi class
        # labels = torch.as_tensor(labels, dtype=torch.int64)
        labels = torch.ones((num_objs,), dtype=torch.int64)

        masks = torch.as_tensor(masks, dtype=torch.uint8)

        image_id = torch.tensor([idx])
        area = (boxes[:, 3] - boxes[:, 1]) * (boxes[:, 2] - boxes[:, 0])
        # suppose all instances are not crowd
        iscrowd = torch.zeros((num_objs,), dtype=torch.int64)

        target = {}
        target["boxes"] = boxes
        target["labels"] = labels
        target["masks"] = masks
        target["image_id"] = image_id
        target["area"] = area
        target["iscrowd"] = iscrowd

        if self.tensor_transforms is not None:
            img, target = self.tensor_transforms(img, target)

        return img, target

    def __len__(self):
        return len(self.img_paths)

# Cell

class ToTensor(object):
    """ Transforms an object (image) into a Tensor
    """
    def __call__(self, image, target):
        image = F.to_tensor(image)
        return image, target

# Internal Cell

def _get_tensor_transforms(train: bool):
    return ToTensor()


def _get_instance_segmentation_dataset(
    *,
    get_tensor_transforms: Callable[
        [bool], Callable[[Image.Image], Any]
    ] = _get_tensor_transforms,
    batch_size: int = 4,
    num_workers: int = 4,
    n_samples: int=-1,
) -> Tuple[
    torch.utils.data.dataloader.DataLoader, torch.utils.data.dataloader.DataLoader
]:
    """Get dataset for instance segmentation. Make sure you define get_transform function."""

    # get data if needed
    _download_data_if_needed()
    root_path = Path(dataset_root)
    assert root_path.exists()
    assert root_path.is_dir()
    assert len(list(root_path.glob("**/*"))) >= 600

    # use our dataset and defined transformations
    dataset = DolphinsInstanceSegmentationDataset(
        dataset_root / "Train",
        tensor_transforms=get_tensor_transforms(train=True),
        n_samples=n_samples
    )
    dataset_test = DolphinsInstanceSegmentationDataset(
        dataset_root / "Val",
        tensor_transforms=get_tensor_transforms(train=False),
        n_samples=n_samples
    )

    # define training and validation data loaders
    data_loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        collate_fn=utils.collate_fn,
    )

    data_loader_test = torch.utils.data.DataLoader(
        dataset_test,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        collate_fn=utils.collate_fn,
    )

    return data_loader, data_loader_test

# Cell

def stack_imgs(imgs: List[PIL.Image.Image], width: int=None) -> PIL.Image:
    """ Stacks images horizontaly in one large image. Very useful for debugging purposes.
    """
    height = max([img.size[1] for img in imgs])
    imgs = [np.array(img.resize((img.size[0], height))) for img in imgs]
    imgs = np.hstack(imgs)
    img = Image.fromarray(imgs)
    if width:
        height = int(img.size[1] * width / img.size[0])
        img = img.resize((width, height))
    return img

# Cell

def display_batches(data_loader: torch.utils.data.DataLoader, *, n_batches: int=1, width:int=800, show_y: bool=False):
    """ Displays `n_batches`, one batch per row.
    """
    to_pil_img = ToPILImage()
    for i, (x, y) in enumerate(data_loader):
        if i >= n_batches:
            return
        if isinstance(x[0], torch.Tensor):
            x = [to_pil_img(t) for t in x]
        display(stack_imgs(x, width=width))
        if show_y:
            display(y)

# Cell

def get_image2tensor_transforms(train: bool):
    """ Converts image to tensor
    """
    return ToTensor()

def get_dataset(
    name: str,
    *,
    get_tensor_transforms: Callable[[bool], Callable[[Image.Image], Any]] = get_image2tensor_transforms,
    batch_size: int = 4,
    num_workers: int = 4,
    n_samples: int=-1,
) -> Tuple[
    torch.utils.data.dataloader.DataLoader, torch.utils.data.dataloader.DataLoader
]:
    """Get one of two datasets available. The parameter `name` can be one of 'segmentation' and 'classification'"""

    assert name in [
        "segmentation",
        "classification",
    ], f"name should be either 'segmentation' or 'classification', but it is '{name}'."

    if name == "segmentation":
        return _get_instance_segmentation_dataset(
            get_tensor_transforms=get_tensor_transforms,
            batch_size=batch_size,
            num_workers=num_workers,
            n_samples=n_samples,
        )
    elif name == "classification":
        raise NotImplementedError()

# Internal Cell

def _flip_coco_person_keypoints(kps, width):
    flip_inds = [0, 2, 1, 4, 3, 6, 5, 8, 7, 10, 9, 12, 11, 14, 13, 16, 15]
    flipped_data = kps[:, flip_inds]
    flipped_data[..., 0] = width - flipped_data[..., 0]
    # Maintain COCO convention that if visibility == 0, then x, y = 0
    inds = flipped_data[..., 2] == 0
    flipped_data[inds] = 0
    return flipped_data


# Cell

class Compose(object):
    """ Compose a list of transformations into one transformation
    """
    def __init__(self, transforms: List):
        self.transforms = transforms

    def __call__(self, image, target):
        for t in self.transforms:
            image, target = t(image, target)
        return image, target


class RandomHorizontalFlip(object):
    """ Randomly flips image horizontally
    """
    def __init__(self, prob):
        self.prob = prob

    def __call__(self, image, target):
        if random.random() < self.prob:
            height, width = image.shape[-2:]
            image = image.flip(-1)
            bbox = target["boxes"]
            bbox[:, [0, 2]] = width - bbox[:, [2, 0]]
            target["boxes"] = bbox
            if "masks" in target:
                target["masks"] = target["masks"].flip(-1)
            if "keypoints" in target:
                keypoints = target["keypoints"]
                keypoints = _flip_coco_person_keypoints(keypoints, width)
                target["keypoints"] = keypoints
        return image, target



# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/03_Submissions.ipynb (unless otherwise specified).

__all__ = ['submit_model']

# Cell

from pathlib import Path

# Internal Cell

import json
import requests
import base64
from datetime import datetime
import boto3


import tempfile
from zipfile import ZipFile
import torch
import csv
import pandas as pd

from .datasets import get_dataset
from .instance_segmentation.model import iou_metric

# Internal Cell


def get_presigned_url_from_aws(
    path: Path,
    gateway_url: str = "https://1nxt4cd0y2.execute-api.eu-central-1.amazonaws.com/v1/submission",
):
    r = requests.post(gateway_url, params={"filename": path.name})

    assert r.status_code < 300, f"ERROR ({r.status_code}): {r.text}"

    presigned_url = json.loads(r.text)

    return presigned_url

# Internal Cell

def upload_with_presigned_url(
    presigned_url,
    path,
):
    files = {"file": open(path, "rb")}
    r = requests.post(presigned_url["url"], files=files, data=presigned_url["fields"])

    assert r.status_code < 300, f"ERROR: {r.text}"
#     print(f"OK: {r.text}")

# Cell


def submit_model(model, *, alias: str, name: str, email: str) -> None:
    """Run evaluation on the model using validation dataset and submits the result under name/email to the central server.

    The leaderbord will be updated every few days because it requires all submitted models to be veried and tested independantly.
    """

    if alias == "dupin123":
        raise ValueError("Please change default alias to your own")

    if email == "name.surname@gmail.com":
        raise ValueError("Please change default email to your own")

    # evaluate model and save metrics
    _, data_loader_test = get_dataset("segmentation", batch_size=4)
    iou, iou_df = iou_metric(model, data_loader_test.dataset)

    # submission info
    info = pd.Series(
        dict(alias=alias, name=name, email=email, iou=iou),
        name="info",
    ).to_frame()

    # paths
    ts = datetime.now().isoformat()
    submission_name = f"submission-iou={iou:.5f}-{alias}-{email}-{ts}"
    d = Path(submission_name)
    d.mkdir(exist_ok=True, parents=True)

    model_save_path = d / "model.pt"
    metrics_save_path = d / "metrics.csv"
    info_save_path = d / "info.csv"

    # save everything
    torch.save(model, model_save_path)
    iou_df.to_csv(metrics_save_path)
    info.to_csv(info_save_path)

    # zip it
    zip_fname = Path(submission_name + ".zip")
    with ZipFile(zip_fname, "w") as myzip:
        for f in d.glob("*"):
            myzip.write(f)

    #
    presigned_url = get_presigned_url_from_aws(
        path=zip_fname,
        gateway_url="https://1nxt4cd0y2.execute-api.eu-central-1.amazonaws.com/v1/submission",
    )

    upload_with_presigned_url(presigned_url, zip_fname)

    return zip_fname


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/97_Helpers.ipynb (unless otherwise specified).

__all__ = ['ensure', 'ensure_equal', 'ensure_exists', 'fix_readme']

# Cell

from pathlib import Path
import re

# Internal Cell

from fastcore.script import *

# Cell

def ensure(p, msg: str=None):
    """ Ensures that p value is True
    """
    if not p:
        raise ValueError(msg)

def ensure_equal(x, y, msg: str=None):
    """ Ensures that two objects are equal
    """

    if msg is None:
        msg = f"{x} != {y}"
    ensure(x == y, msg)

def ensure_exists(f: Path, msg: str=None):
    """ Ensures that path exists
    """

    if msg is None:
        msg = f"{f} does not exist"
    ensure(f.exists(), msg)

# Cell


@call_parse
def fix_readme(
    file_path: Param("path to README.md file", Path), verbose: Param("", bool) = False
):

    file = open(file_path, "r")
    list_of_lines = file.readlines()
    for i in range(len(list_of_lines)):
        r1 = re.search("{% include note.html content='(.*?)' %}", list_of_lines[i])
        if r1:
            list_of_lines[i] = r1.group(1) + "\n"
            if verbose:
                print(list_of_lines[i])

    file = open(file_path, "w")
    file.writelines(list_of_lines)
    file.close()


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/98_Convert_raw_image_to_jpg.ipynb (unless otherwise specified).

__all__ = ['get_suffixes', 'convert_files_with_darktable']

# Cell


from pathlib import Path
from typing import Tuple, Optional, Dict, List, Union, Callable, Iterable, Set

# Internal Cell


from fastcore.script import *
import numpy as np
from shutil import copy
import os
import random
import time
import progressbar
from progressbar import FormatLabel

# Internal Cell


def glob_suffixes(root_path: Path, suffixes: Union[List[str], str]) -> List[Path]:
    if isinstance(suffixes, str):
        suffixes = [suffixes]
    return sorted([f for f in root_path.glob("**/*") if file_suffix_in(f, suffixes)])

# Cell


@call_parse
def get_suffixes(
    path: Param("input directory containing images", Path),
    verbose: Param("", bool) = True,
) -> Set[str]:
    """Returns all unique suffixes"""
    path = Path(path)
    p = path.glob("**/*")
    files = [x.suffix for x in p if x.is_file()]
    if verbose:
        print(list(set(x.lower() for x in files)))
    return set(x.lower() for x in files)

# Internal Cell


def get_rand_file_with_suff(
    path: Path, suffix: str, case_insensitive: bool = True
) -> Path:
    """Returns a random file that has a suffix equal to the variable suff"""
    if case_insensitive:
        p = [p for p in path.glob(f"**/*") if p.suffix.lower() == suffix.lower()]
    else:
        p = path.glob(f"**/*{suffix}")
    files = [x for x in p if x.is_file()]
    return random.choice(files)

# Internal Cell


def get_one_image_for_each_suffix(path: Path) -> List[Path]:
    """Returns one random image for each suffix"""
    return [
        get_rand_file_with_suff(path, suffix)
        for suffix in get_suffixes(path, verbose=False)
    ]

# Internal Cell


def create_sample_test_dir_if_needed(
    dst: Path = Path("../../data/testing/raw"),
) -> List[Path]:
    """Creates samples for a test set"""
    dst.mkdir(parents=True, exist_ok=True)
    if len(list(dst.glob("*"))) == 0:
        files = get_one_image_for_each_suffix(dropbox_path)
        for f in files:
            copy(f, dst)
    return list(dst.glob("*"))

# Internal Cell


def convert_image_to_jpg_darktable(
    img_path: Path,
    *,
    suffixes_to_delete: List[str] = [".mov", ".mp4"],
    remove_src: bool = False,
    update_msg_fn: Callable[[str], None] = print,
) -> str:
    update_msg_fn(f"Processing {img_path}")
    """Converts image whose suffix is not in the skip_suffixes list, if remove_src = True then deletes the input image"""
    if img_path.suffix.lower() != ".jpg":
        dst = img_path.parent / (img_path.stem + ".jpg")
        if img_path.suffix.lower() in suffixes_to_delete:
            if remove_src:
                retmsg = f"Removed {img_path} because {img_path.suffix.lower()} is in {suffixes_to_delete}"
                img_path.unlink()
            else:
                retmsg = f"Not removing {img_path} although {img_path.suffix.lower()} is in {suffixes_to_delete}"
            return None, retmsg
        if dst.exists() and remove_src:
            img_path.unlink()
            retmsg = f"Removed {img_path} because {dst} exists."
            return dst, retmsg
        system_error = os.system(f"darktable-cli {img_path} {dst}")
        if system_error != 0:
            retmsg = f"Converted error {system_error}, image path {img_path}."
            return img_path, retmsg
        if remove_src:
            img_path.unlink()
        retmsg = f"Converted {img_path} to {dst}."
        return dst, retmsg
    else:
        retmsg = f"Skiped path {img_path} because it is already a JPEG."
        return img_path, retmsg

# Cell


@call_parse
def convert_files_with_darktable(
    root_path: Param("input directory containing images to convert", Path),
    remove_src: Param("input True if you want to delete the images ", bool) = True,
):
    """Calls the convert_image_to_jpg() function on all images in the path"""
    root_path = Path(root_path)
    widgets = [
        FormatLabel(""),
        " [",
        progressbar.Timer(),
        "] ",
        progressbar.Bar(),
        progressbar.Percentage("%(percentage)5.2f%%"),
        " (",
        progressbar.AdaptiveETA(),
        ") ",
    ]
    files = sorted([x for x in root_path.glob("**/*") if x.is_file()])
    n_files = len(files)

    def update_msg(msg):
        widgets[0] = FormatLabel(f"[{i:4d}/{n_files}] {msg}")

    log = []
    with progressbar.ProgressBar(max_value=len(files), widgets=widgets) as bar:
        for i, src in enumerate(files):
            _, msg = convert_image_to_jpg_darktable(
                src, remove_src=remove_src, update_msg_fn=update_msg
            )
            log.append(msg)
            bar.update(i)
    print(f"Conversion completed {len(files)}/{len(files)}")

    for msg in log:
        print(msg)